import logging
import numpy as np
import cv2
import logging
from PIL import Image
from os.path import splitext
from os import listdir
from torch.utils.data import Dataset
import torch


import seg_deeplab as network


class BasicDataset(Dataset):
    def __init__(self, folder, transform):
        self.imgs_dir = f"{folder}/imgs/"
        self.masks_dir = f"{folder}/masks/"
        self.transform = transform

        self.ids = [splitext(file)[0] for file in listdir(self.imgs_dir) if not file.startswith(".")]
        logging.info(f"Creating dataset with {len(self.ids)} examples")

    def __len__(self):
        return len(self.ids)

    @classmethod
    def pre_process(self, img):
        if len(img.shape) == 2:
            img = np.expand_dims(img, axis=2)
        img = img.transpose((2, 0, 1))
        if img.max() > 1:
            img = img / 255
        return img

    def __getitem__(self, i):
        idx = self.ids[i]
        mask_file = f"{self.masks_dir}{idx}.png"
        img_file = f"{self.imgs_dir}{idx}.png"

        img = np.array(Image.open(img_file).convert("RGB"))
        mask = np.array(Image.open(mask_file).convert("L"))

        data = {"image": img, "mask": mask}
        augmented = self.transform(**data)
        img, mask = augmented["image"], augmented["mask"]

        # cv2.imshow("img", img)
        # cv2.waitKey(0)

        # HWC to CHW
        img = self.pre_process(img)
        mask = self.pre_process(mask)

        mask[mask >= 0.5] = 1
        mask[mask < 0.5] = 0

        return torch.from_numpy(img).type(torch.FloatTensor), torch.from_numpy(mask).type(torch.FloatTensor)


model_map = {
    "deeplabv3_resnet50": network.deeplabv3_resnet50,
    "deeplabv3plus_resnet50": network.deeplabv3plus_resnet50,
    "deeplabv3_resnet101": network.deeplabv3_resnet101,
    "deeplabv3plus_resnet101": network.deeplabv3plus_resnet101,
    "deeplabv3_mobilenet": network.deeplabv3_mobilenet,
    "deeplabv3plus_mobilenet": network.deeplabv3plus_mobilenet,
}


class SegNet:
    def __init__(self, model_name, checkpoint_path, img_w, img_h):
        self.model = model_map[model_name](num_classes=1, output_stride=16)
        network.convert_to_separable_conv(self.model.classifier)

        logging.info("Loading model {}".format(model_name))

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        logging.info(f"Using device {self.device}")

        checkpoint = torch.load(checkpoint_path, map_location=torch.device("cpu"))
        self.model.load_state_dict(checkpoint)
        self.model.to(self.device)

        logging.info("Model loaded !")

        self.img_w = img_w
        self.img_h = img_h

    def predict_img(self, img):
        self.model.eval()

        img = cv2.resize(img, (self.img_w, self.img_h))

        img = torch.from_numpy(BasicDataset.pre_process(np.array(img)))

        img = img.unsqueeze(0)
        img = img.to(device=self.device, dtype=torch.float32)

        with torch.no_grad():
            output = self.model(img)

            probs = torch.sigmoid(output)
            probs = probs.squeeze(0).cpu()

            full_mask = probs.squeeze().cpu().numpy()

        result = full_mask / np.max(full_mask)
        result = (result * 255).astype(np.uint8)
        return result
